;> File name: FBUT.SC
;>
;>  Build library containing procs to create and manage floating
;>  BUTTON, CHECKBOX, and RADIOBUTTON windows.
;\

message "Creating lib FBUT, Please wait .."
createlib "fbut"
release procs wproc
 proc wproc(x)
    execute "writelib \"fbut\" " + x
    + " release procs " + x
 endproc

;> The the interface procs in script FBUTI are included
;> in the lib.
;\
play "FBUTI"

;=================================================== FBUT_Create
proc FBUT_Create
; Create a floating BUTTON, RADIOBUTTON, or CHECKBOX window.
;
; The window is invisible and inactive when it is created.
;
; Call FBUT_Visible to make it visible (or invisible)
; Call FBUT_Active to make it active (or inactive)
; Call FBUT_Move to change its location or parent.
;
; Normally, this proc is not called directly. You should set
; up some procs that accept normal parameter lists, build
; the lists (dynarrays) required by this proc, and then call
; this proc. Your "interface" procs can be written to choose
; defaults that are appropriate to your application.
;
;
; The indexes used in dynarray wininfo are described below.
;
; Default values will be supplied if no value is provided by
; the caller.
;
; Index        Description
; ------------ ----------------------------------------------------
; PARENT       Parent window handle. The floating window's origin
;              is relative to the parent's origin.
;              Call FBUT_Visible after the parent is moved or
;              resized to realign the floating window with
;              its parent.
;
; ROW          The origin row of the window.
;
; COL          The origin col of the window.
;
; HEIGHT       The height of the window.
;              The value will be increased if the labels wil not
;              fit in the window.
;              Use 0 (zero) for autosizing of the window.
;
; WIDTH        The width of the window.
;              The value will be increased if the labels wil not
;              fit in the window.
;              Use 0 (zero) for autosizing of the window.
;
; HIDEROW      The origin row of the window when it is not visible.
;              Defaults to -200.
;
; HIDECOL      The origin col of the window when it is not visible.
;              Defaults to 0.
;
; TYPE         Type of floating window. The following 1-char codes
;              are used:
;              B=BUTTON, R=RADIOBUTTON, or C=CHECKBOX.
;
; STYLE        Integer. Window style within TYPE.
;              The supported styles are described below.
;
;              Type  Style Description
;              ----  ----- ----------------------------------------
;               B      1   No frame. No shadow. Height = 1.
;               R      1   No frame. No shadow.
;               C      1   No frame. No shadow.
;
;             This parameter can be used to support new window
;             styles. One style that I tried and discarded was
;             BUTTON style 2 for a button with an internal frame:
;
;                             Ŀ
;                                 Ok    
;                             
;
;             This example should give you a pretty good idea of how
;             to use the style parameter.
;
; NUMCOL      THIS PARAMETER IS IGNORED. The original intent was to
;             support multi-column checkboxes and radiobuttons. I
;             never needed them, so I didn't put any multi-column
;             support in the Paint and Draw procs.
;
;             However, most of the code was written with multi-
;             column support in mind and should support
;             numcol-column radiobuttons and checkboxes if
;             appropriate modifications are made to the Paint and
;             Draw procs.
;
; HEADING     The heading for a RADIOBUTTON or CHECKBOX window.
;             The value is ignored for a BUTTON window.
;             The heading optional. If it is specified, it must
;             fit on a single line in the window.
;
; VARNAME     ** This is a string that contains a var name. **
;             It names the variable to which a value is assigned
;             when a label in the floating window is "selected".
;
;             If a var name is associated with a BUTTON window,
;             then the var is set equal to the button's label
;             (with tildes removed).
;
;             If a var name is assigned to a RADIOBUTTON window,
;             then the var is set to the index of the selected
;             label.
;
;             If a var name is assigned to a CHECKBOX, then
;             the var must be a dynarray. If label #n in the
;             box is selected (checked), then element [n] of
;             the dynarray is TRUE. Otherwise, element [n]
;             is set to FALSE.
;
;             Regardless of the type of box, the named var must
;             be defined before its associated box is created.
;
; PROCNAME    The name of the proc to EXECPROC when a label in
;             the window is selected. The proc will be invoked
;             after the VARNAME var (if any) has been updated
;             but before the window is updated to reflect the
;             new value of the var.
;
; ACTIVE      New windows are set as "inactive" when they are
;             created. Set this element to TRUE if you want the
;             new window to be automatically activated.
;             (See proc FBUT_Active for more information.)
;
; VISIBLE     New windows are invisible when they are created.
;             Set this element to TRUE if you want the new
;             window to be visible.
;             (See proc FBUT_Visible for more information.)
;
; The following indexes may be added to wininfo.
;
; Index        Description
; ------------ ----------------------------------------------------
; NUMLBL       The number of labels defined in x.label.
; MAXWID       The width of the widest label.
; HASHEADING   TRUE if the window has a heading. Otherwise, FALSE.
;
;
; The indexes used in dynarray x.label are described below.
;
; Note: x.label may also be specified as a newline-separated (\n)
;       list of labels. DO NOT put a newline after the last label.
;
;
; Index        Description
; ------------ ----------------------------------------------------
; n            n is the label number. It starts at 1 and increases
;              by 1 for each successive label. There can be NO
;              gaps in the numbering.
;
;              The element contains the text of the nth label.
;
;              Use paired tildes ~~ to delimit a highlighted
;              substring within the label. Or specify the start
;              position for highlighting in x.hipos and the length
;              in x.hilen.
;
;              If no tildes are found and the corresponding
;              entry in x.hipos is not assigned, then the
;              first char in the label is highlighted.
;
;              A BUTTON window uses only label 1.
;
;
; The indexes used in dynarray x.keycode are described below.
;
; Note: x.keycode may also be specified as a comma-separated
;       list of keycodes.
;
; Index        Description
; ------------ ----------------------------------------------------
; n            Uses the same indexes as x.label.
;
;              The element contains the keycode of the hotkey for
;              the corresponding label in x.label.
;
;              If keycode is 0 or no dynarray entry is supplied,
;              then the Alt key value of the first highlighted
;              character in the corresponding label is used as the
;              keycode. E.g., if the first highlighted char in a
;              label is "A" or "a", then Alt-A will be used as the
;              hotkey.
;
;              Specify "" or BLANKNUM() to specify that the
;              corresponding label has no hotkey. A label
;              that has no hotkey can only be selected by using
;              a mouse.
;
; The indexes used in dynarray x.hipos and x.hilen are
; described below.
;
; Index        Description
; ------------ ----------------------------------------------------
; n            Uses the same indexes as x.label.
;
;              The x.hipos element contains the start position for
;              a substring that is to be highlighted.
;              The start position is relative to 1, i.e., the first
;              character in the label is at position 1.
;
;              The x.hilen element contains the length of the
;              highlighted substring.
;
;              To suppress highlighting for a label, specify 0
;              (zero) as the start position in x.hipos or zero
;              as the length in x.hilen. If either is zero, both
;              will be set to zero.
;
;              If either one is specified (x.hipos or x.hilen)
;              for a label, and the other is not specified or set
;              to zero automatically, then the specification is
;              ignored and the label will be examined for tildes.
;
;              If x.hilen and x.hipos are both specified, then the
;              label will not be examined for tildes.
;
;              Normally, x.hipos and x.hilen are not set by the
;              caller. They are derived by examining the label
;              for tildes.
;
;
; Return value:
; The window's handle. Zero if the window is not created.
; The content of the dynarrays in the parameter list may be
; modified.
(
; ******** The following parameters are described above ***********

    wininfo       ; {I,O} Dynarray of formatting information.
  , x.label       ; {I,O} Dynarray of labels.
  , x.keycode     ; {I,O} Dynarray of hotkey keycodes for labels.
  , x.hipos       ; {I,O} Dynarray of hilite starting positions.
  , x.hilen       ; {I,O} Dynarray of hilite lengths.
)
private x, y, z
    , orig.canv
    , orig.win
    , label.txt
    , save.active
    , save.visible
    , winattr

; Save current canvas and window. The window won't be
; correct unless the caller has done a WINDOW SELECT
; for the form or image.

orig.canv = getcanvas()
orig.win = getwindow()

; Make sure global vars are initialized
if not isassigned( fbut.globals.are.initialized ) then
    FBUT_Init_Globals()
endif


; Set defaults for all wininfo elements. This simplifies
; the remainder of the code (we do not have to check for
; existence).

if isassigned( wininfo[ "parent" ] ) then
    if not iswindow( wininfo[ "parent" ] ) then
        wininfo[ "parent" ] = 0
    endif
else
    wininfo[ "parent" ] = 0
endif

if not isassigned( wininfo[ "row" ] ) then
    wininfo[ "row" ] = 0
endif

if not isassigned( wininfo[ "col" ] ) then
    wininfo[ "col" ] = 0
endif

if not isassigned( wininfo[ "hiderow" ] ) then
    wininfo[ "hiderow" ] = -200
endif

if not isassigned( wininfo[ "hidecol" ] ) then
    wininfo[ "hidecol" ] = 0
endif

if not isassigned( wininfo[ "height" ] ) then
    wininfo[ "height" ] = 1
endif

if not isassigned( wininfo[ "width" ] ) then
    wininfo[ "width" ] = 1
endif

if not isassigned( wininfo[ "type" ] ) then
    wininfo[ "type" ] = "B"  ; assume button
endif

wininfo[ "type" ] = upper( wininfo[ "type" ] )

if len( wininfo[ "type" ] ) > 1 then
    wininfo[ "type" ] = substr( wininfo[ "type" ], 1, 1 )
endif

if search( wininfo[ "type" ], "BRC" ) = 0 then
    wininfo[ "type" ] = "B"  ; force to button type
endif

if not isassigned( wininfo[ "style" ] ) then
    wininfo[ "style" ] = 1
endif

if not isassigned( wininfo[ "heading" ] ) then
    wininfo[ "heading" ] = ""
endif

wininfo[ "hasheading" ] = iif( isblank( wininfo[ "heading" ] ), false, true )

if not isassigned( wininfo[ "varname" ] ) then
    wininfo[ "varname" ] = ""
endif

if not isassigned( wininfo[ "procname" ] ) then
    wininfo[ "procname" ] = ""
endif

if not isassigned( wininfo[ "active" ] ) then
    wininfo[ "active" ] = false
endif

if not isassigned( wininfo[ "visible" ] ) then
    wininfo[ "visible" ] = false
endif

; x.label may be a newline-separated list of labels
if match( type( x.label ), "A.." ) or
   match( type( x.label ), "M.." ) then
    x = x.label
    dynarray x.label[]
    FBUT_Parse_List( x, x.label, "\n" )
endif

; Verify that the labels are in a dynarray with no
; gaps in the numbering.

if type( x.label ) = "DY" then
    for x from 1 to dynarraysize( x.label )
        if not isassigned( x.label[ x ] ) then
            x.label[ x ] = ""
        endif
    endfor
else
    dynarray x.label[]
    x.label[ 1 ] = "Invalid parameter type for x.label"
endif

if dynarraysize( x.label ) = 0 then
    x.label[ 1 ] = "NO LABEL SPECIFIED"
    x.hipos[ 1 ] = 0
    x.hilen[ 1 ] = 0
endif

wininfo[ "numlbl" ] = dynarraysize( x.label )

; x.keycode may be a comma-separated list of keycodes

if match( type( x.keycode ), "A.." ) or
   match( type( x.keycode ), "M.." ) then
    x = x.keycode
    dynarray x.keycode[]
    if not isblank( x ) then
        FBUT_Parse_List( x, x.keycode, "," )
    endif
endif

; Verify that the keycodes are numbers.
;
; We only care about keycodes that correspond to
; labels.

if type( x.keycode ) = "DY" then
    for x from 1 to wininfo[ "numlbl" ]
        if isassigned( x.keycode[ x ] ) then
            if isblank( x.keycode[ x ] ) then
                x.keycode[ x ] = blanknum()
            endif
            if type( x.keycode[ x ] ) <> "N" then
                x.keycode[ x ] = numval( x.keycode[ x ] )
            endif
            if type( x.keycode[ x ] ) <> "N" then
                release vars x.keycode[ x ]
            endif
        endif
    endfor
else
    dynarray x.keycode[]
endif

; Verify that highlight positioning entries are in dynarrays
; and are numbers that make sense.
;
; We only care about entries that correspond to labels.

if type( x.hipos ) = "DY" and type( x.hilen ) = "DY" then
    for x from 1 to wininfo[ "numlbl" ]
        if isassigned( x.hipos[ x ] ) then
            if isblank( x.hipos[ x ] ) then
                x.hipos[ x ] = blanknum()
                x.hilen[ x ] = blanknum()
            endif
            if type( x.hipos[ x ] ) <> "N" then
                x.hipos[ x ] = numval( x.hipos[ x ] )
            endif
            if type( x.hipos[ x ] ) <> "N" then
                release vars x.hipos[ x ], x.hilen[ x ]
            endif
            if not isassigned( x.hilen[ x ] ) then
                release vars x.hipos[ x ]
            endif
        endif
        if isassigned( x.hilen[ x ] ) then
            if isblank( x.hilen[ x ] ) then
                x.hipos[ x ] = blanknum()
                x.hilen[ x ] = blanknum()
            endif
            if type( x.hilen[ x ] ) <> "N" then
                x.hilen[ x ] = numval( x.hilen[ x ] )
            endif
            if type( x.hilen[ x ] ) <> "N" then
                release vars x.hipos[ x ], x.hilen[ x ]
            endif
            if not isassigned( x.hipos[ x ] ) then
                release vars x.hilen[ x ]
            endif
        endif
        if isassigned( x.hipos[ x ] ) then
            if x.hipos[ x ] = 0 then
                x.hilen[ x ] = 0
            endif
            if x.hilen[ x ] = 0 then
                x.hipos[ x ] = 0
            endif
            if x.hipos[ x ] < 0 or   ; Gets rid of blanknum(), too
               x.hilen[ x ] < 0 or
               x.hipos[ x ] + x.hilen[ x ] - 1 > len( x.label[ x ] ) then
                release vars x.hipos[ x ], x.hilen[ x ]
            endif
        endif
    endfor
else
    dynarray x.hipos[]
    dynarray x.hilen[]
endif

; Derive highlight and hotkey info

for x from 1 to wininfo[ "numlbl" ]
    if not isassigned( x.hipos[ x ] ) then
        label.txt = x.label[ x ]
        ; Remove leading and trailing spaces
        while match( label.txt, ".. ", label.txt ) endwhile
        while match( label.txt, " ..", label.txt ) endwhile
        if not isblank( label.txt ) then
            dynarray z[]
            x.label[ x ] = FBUT_Extract_Delimiters( label.txt, "", z )
            if z[ "length" ] = 0 then
                x.hipos[ x ] = 1
                x.hilen[ x ] = 1
            else
                x.hipos[ x ] = z[ "startpos" ]
                x.hilen[ x ] = z[ "length" ]
            endif
        else
            x.label[ x ] = ""
            x.hipos[ x ] = 0
            x.hilen[ x ] = 0
        endif
    endif
    if not isassigned( x.keycode[ x ] ) or
       x.keycode[ x ] = 0 then
        if x.hipos[ x ] > 0 then
            z = substr( x.label[ x ], x.hipos[ x ], 1 )
            x.keycode[ x ] = FBUT_Get_Alt_Keycode( z )
        else
            x.keycode[ x ] = blanknum()
        endif
    endif
endfor

; Put the length of the longest label into wininfo[ "maxwid" ]
z = 0
for x from 1 to wininfo[ "numlbl" ]
    if isblank( x.label[ x ] ) then
        x.label[ x ] = " "  ; need len > 0
    endif
    z = max( z, len( x.label[ x ] ) )
endfor
wininfo[ "maxwid" ] = z

execproc fbut.draw.proc[ wininfo[ "type" ] ]
winhand = retval

if iswindow( winhand ) then
    ; Register new window.
    ;
    ; This saves info about the window in global dynarrays.

    save.visible = wininfo[ "visible" ]  ; not visible yet
    wininfo[ "visible" ] = false
    save.active = wininfo[ "active" ]    ; not active yet
    wininfo[ "active" ] = false

    FBUT_Register_New_Window( winhand, wininfo )

    ; Get window attributes for Paint and Update
    window getattributes winhand to winattr

    ; Paint and Update use the info saved by Register
    execproc fbut.paint.proc[ wininfo[ "type" ] ]

    ; Update uses info saved by Paint
    execproc fbut.update.proc[ wininfo[ "type" ] ]

    if save.active then
        FBUT_Active( winhand, true )
    endif

    if save.visible then
        FBUT_Visible( winhand, true )
    endif
endif

if iswindow( orig.canv ) then
    setcanvas orig.canv
else
    setcanvas default
endif

if iswindow( orig.win ) then
    window select orig.win
endif

return winhand

endproc
wproc( "FBUT_Create" )

;=================================================== FBUT_Active
proc FBUT_Active
; Make a floating window active or inactive.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
  , x.active      ; {I} TRUE->make window active.
                  ;     FALSE->make window inactive.
)
private x, y, z
    , wininfo

; Make sure the window exists
if iswindow( winhand ) then
    dynarray wininfo[]
    if FBUT_Get_Window_Info( winhand, wininfo ) then
        if wininfo[ "active" ] <> x.active then
            wininfo[ "active" ] = x.active
            ; Activate or deactivate the labels' hotkeys
            for x from 1 to wininfo[ "numlbl" ]
                FBUT_Hotkey_Active( winhand, x, x.active )
            endfor
            FBUT_Set_Window_Info( winhand, wininfo )
            FBUT_Paint( winhand )
        endif
    endif
endif

endproc
wproc( "FBUT_Active" )

;=================================================== FBUT_Visible
proc FBUT_Visible
; Make a floating window visible or invisible.
;
; If the window is visible, its screen coordinates will be
; adjusted to align with the parent window, if any.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
  , x.visible     ; {I} TRUE->make window visible.
                  ;     FALSE->make window invisible.
)
private x, y, z
    , wininfo
    , par.row
    , par.col


; Make sure the window exists
if iswindow( winhand ) then
    dynarray wininfo[]
    if FBUT_Get_Window_Info( winhand, wininfo ) then
        if x.visible then
            ; The parent window may have moved. We can't
            ; avoid checking coordinates.
            if iswindow( wininfo[ "parent" ] ) then
                window getattributes wininfo[ "parent" ] to x
                par.row = x[ "originrow" ]
                par.col = x[ "origincol" ]
                if x[ "hasframe" ] then
                    par.row = par.row + 1
                    par.col = par.col + 1
                endif
            else
                wininfo[ "parent" ] = 0
                par.row = 0
                par.col = 0
            endif
            wininfo[ "visible" ] = true
            dynarray x[]
            x[ "originrow" ] = par.row + wininfo[ "row" ]
            x[ "origincol" ] = par.col + wininfo[ "col" ]
            window setattributes winhand from x
            FBUT_Set_Window_Info( winhand, wininfo )
        else
            if wininfo[ "visible" ] then
                wininfo[ "visible" ] = false
                dynarray x[]
                x[ "originrow" ] = wininfo[ "hiderow" ]
                x[ "origincol" ] = wininfo[ "hidecol" ]
                window setattributes winhand from x
                FBUT_Set_Window_Info( winhand, wininfo )
            else
                ; already invisible - nothing to do
            endif
        endif
    endif
endif

endproc
wproc( "FBUT_Visible" )

;=================================================== FBUT_Move
proc FBUT_Move
; Move a window and/or assign a new parent window.
;
; If the window is visible, its screen coordinates will be
; adjusted to align with the parent window, if any.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
  , x.parent      ; {I} The window handle of the parent window.
                  ;     0 (zero) to set to "no parent"
                  ;     Blank ("" or blanknum()) to use same parent.
  , x.row         ; {I} Origin row of window (Relative to parent)
                  ;     Blank ("" or blanknum()) to use same row
  , x.col         ; {I} Origin col of window (Relative to parent)
                  ;     Blank ("" or blanknum()) to use same col
)
private x, y, z
    , wininfo
    , x.visible

; Make sure the window exists
if iswindow( winhand ) then
    dynarray wininfo[]
    if FBUT_Get_Window_Info( winhand, wininfo ) then
        x.visible = wininfo[ "visible" ]
        if not isblank( x.row ) then
            wininfo[ "row" ] = x.row
        endif
        if not isblank( x.col ) then
            wininfo[ "col" ] = x.col
        endif
        if not isblank( x.parent ) then
            wininfo[ "parent" ] = x.parent
        endif
        FBUT_Set_Window_Info( winhand, wininfo )
        if x.visible then
            FBUT_Visible( winhand, true )
        endif
    endif
endif

endproc
wproc( "FBUT_Move" )

;=================================================== FBUT_Destroy
proc FBUT_Destroy
; Close a floating button, radiobutton, or checkbox window.
;
; Return value:
; None.
(
    winhand       ; {I} The handle of the window to close.
)
private w, x, y, z
	, orig.canv
	, orig.win
    , hotkey.keycode
    , num.lbl

; Save current canvas and window. The window won't be
; correct unless the caller has done a WINDOW SELECT
; for the form or image.

orig.canv = getcanvas()
orig.win = getwindow()

if iswindow( winhand ) then
    window select winhand
    window close
endif

if isassigned( fbut.window.parent[ winhand ] ) then
    release vars fbut.window.parent[ winhand ]

    x = strval( winhand ) + ":"

    num.lbl = fbut.window.info[ x + "numlbl" ]
    for z from 1 to num.lbl
        y = x + strval( z )  ; winhand : label number
        hotkey.keycode = fbut.label.info[ y + ":keycode" ]
        if isassigned( fbut.hotkey.list[ hotkey.keycode ] ) and
           fbut.hotkey.list[ hotkey.keycode ] = y then
            release vars fbut.hotkey.list[ hotkey.keycode ]
        endif
        y = strval( winhand ) + ":" + strval( z ) + ":"
        for w from 1 to arraysize( fbut.labelreg )
            release vars fbut.label.info[ y + fbut.labelreg[ w ] ]
        endfor
    endfor

    z = strval( winhand ) + ":"
    for x from 1 to arraysize( fbut.winreg )
        if lower( fbut.winreg[ x ] ) = "parent" then
            loop
        endif
        release vars fbut.window.info[ z + fbut.winreg[ x ] ]
    endfor

endif

if iswindow( orig.canv ) then
    setcanvas orig.canv
else
    setcanvas default
endif

if iswindow( orig.win ) then
    window select orig.win
endif

endproc
wproc( "FBUT_Destroy" )

;=================================================== FBUT_Destroy_All
proc FBUT_Destroy_All
; Close all floating button, checkbox, and radiobutton windows.
;
; Return value:
; None.
(
)
private x, y, z
	, orig.canv
	, orig.win

; Save current canvas and window. The window won't be
; correct unless the caller has done a WINDOW SELECT
; for the form or image.

orig.canv = getcanvas()
orig.win = getwindow()

; Close all of the floating BUTTON, etc. windows

foreach x in fbut.window.parent
    z = numval( x )
    if iswindow( z ) then
        window select z
        window close
    endif
endforeach

; Empty all of the global dynarrays

dynarray fbut.label.info[]
dynarray fbut.window.info[]
dynarray fbut.window.parent[]
dynarray fbut.hotkey.list[]

if iswindow( orig.canv ) then
    setcanvas orig.canv
else
    setcanvas default
endif

if iswindow( orig.win ) then
    window select orig.win
endif

endproc
wproc( "FBUT_Destroy_All" )

;=================================================== FBUT_Init_Globals
proc FBUT_Init_Globals
; Initialize global vars used by FBUT_* procs
;
; Return value:
; None but the globals are initialized.
(
)

; The dynarray fbut.window.parent is indexed by window handle and
; an entry exists for every BUTTON, RADIOBUTTON, or CHECKBOX
; floating window. The entry value is the parent window's handle.
;
; Each floating window has a parent window assigned. If there is
; no real parent, then the desktop is considered to be the parent.
; In this case, we use  0 (zero) as the parent window's handle.

dynarray fbut.window.parent[]



; The dynarray fbut.window.info is indexed by window handle and
; field name. The window handle and field name are separated by
; a colon ":". For example: strval( handle ) + ":" + "ROW"
;
; This dynarray contains information about each floating window.
;
; See proc FBUT_Register_New_Window for a description of the
; various fields.

dynarray fbut.window.info[]



; The dynarray fbut.label.info is indexed by window handle,
; label number, and field name. Each index element is separated
; from the others by a colon.
;
; This dynarray contains information about each label in a window.
;
; See proc FBUT_Store_Label_Info for a description of the various
; fields.

dynarray fbut.label.info[]



; The dynarray fbut.hotkey.list is indexed by hotkey keycode. The
; value of the entry is the window handle of the window that
; contains the label associated with the hotkey and the label
; number of the label within the window. These are string values
; separated by a colon.
;
; There is an entry in the list for each hotkey associated with an
; active floating window. Entries for inactive windows are removed
; from the list.

dynarray fbut.hotkey.list[]



; Colors for BUTTON windows

fbut.color.button.window.active        = syscolor( 1040 )
fbut.color.button.hotkey.active        = syscolor( 1044 )
fbut.color.button.window.inactive      = syscolor( 1040 )
fbut.color.button.hotkey.inactive      = syscolor( 1040 )

; Colors for RADIOBUTTON windows

fbut.color.radio.window.active         = syscolor( 1046 )
fbut.color.radio.title.active          = syscolor( 1047 )
fbut.color.radio.hotkey.active         = syscolor( 1048 )
fbut.color.radio.window.inactive       = syscolor( 1046 )
fbut.color.radio.title.inactive        = syscolor( 1046 )
fbut.color.radio.hotkey.inactive       = syscolor( 1046 )

; Colors for CHECKBOX windows

fbut.color.check.window.active         = syscolor( 1046 )
fbut.color.check.title.active          = syscolor( 1047 )
fbut.color.check.hotkey.active         = syscolor( 1048 )
fbut.color.check.window.inactive       = syscolor( 1046 )
fbut.color.check.title.inactive        = syscolor( 1046 )
fbut.color.check.hotkey.inactive       = syscolor( 1046 )

; Proc names

dynarray fbut.draw.proc[]
fbut.draw.proc[ "B" ] = "FBUT_Draw_Button"
fbut.draw.proc[ "R" ] = "FBUT_Draw_Radio"
fbut.draw.proc[ "C" ] = "FBUT_Draw_CheckBox"

dynarray fbut.paint.proc[]
fbut.paint.proc[ "B" ] = "FBUT_Paint_Button"
fbut.paint.proc[ "R" ] = "FBUT_Paint_Radio"
fbut.paint.proc[ "C" ] = "FBUT_Paint_CheckBox"

dynarray fbut.update.proc[]
fbut.update.proc[ "B" ] = "FBUT_Update_Button"
fbut.update.proc[ "R" ] = "FBUT_Update_Radio"
fbut.update.proc[ "C" ] = "FBUT_Update_CheckBox"


; Array that contains the dynarray index names for
; FBUT_Register_New_Window, FBUT_Get_Window_Info,
; FBUT_Set_Window_Info, and one or two other procs.
;
; The indexes are described in FBUT_Create.

array fbut.winreg[ 13 ]
fbut.winreg[  1 ] = "type"
fbut.winreg[  2 ] = "varname"
fbut.winreg[  3 ] = "procname"
fbut.winreg[  4 ] = "parent"
fbut.winreg[  5 ] = "row"
fbut.winreg[  6 ] = "col"
fbut.winreg[  7 ] = "hiderow"
fbut.winreg[  8 ] = "hidecol"
fbut.winreg[  9 ] = "active"
fbut.winreg[ 10 ] = "visible"
fbut.winreg[ 11 ] = "numlbl"
fbut.winreg[ 12 ] = "style"
fbut.winreg[ 13 ] = "hasheading"

; Array that contains dynarray index names for
; FBUT_Store_Label_Info and FBUT_Retrieve_Label_Info.

array fbut.labelreg[ 7 ]
fbut.labelreg[  1 ] = "row"
fbut.labelreg[  2 ] = "col"
fbut.labelreg[  3 ] = "width"
fbut.labelreg[  4 ] = "hotpos"
fbut.labelreg[  5 ] = "hotwid"
fbut.labelreg[  6 ] = "keycode"
fbut.labelreg[  7 ] = "text"



; Dummy var to indicate that globals have been initialized.

;        *** DON'T MODIFY OR RELEASE THIS VAR ****

fbut.globals.are.initialized = true

endproc
wproc( "FBUT_Init_Globals" )

;=================================================== FBUT_Release_Globals
proc FBUT_Release_Globals
; Release the global vars used by FBUT_* procs
;
; Return value:
; None but the globals are released.
(
)

release vars
    fbut.window.parent
  , fbut.window.info
  , fbut.label.info
  , fbut.hotkey.list
  , fbut.color.button.window.active
  , fbut.color.button.hotkey.active
  , fbut.color.button.window.inactive
  , fbut.color.button.hotkey.inactive
  , fbut.color.radio.window.active
  , fbut.color.radio.title.active
  , fbut.color.radio.hotkey.active
  , fbut.color.radio.window.inactive
  , fbut.color.radio.title.inactive
  , fbut.color.radio.hotkey.inactive
  , fbut.color.check.window.active
  , fbut.color.check.title.active
  , fbut.color.check.hotkey.active
  , fbut.color.check.window.inactive
  , fbut.color.check.title.inactive
  , fbut.color.check.hotkey.inactive
  , fbut.draw.proc
  , fbut.paint.proc
  , fbut.update.proc
  , fbut.winreg
  , fbut.labelreg
  , fbut.globals.are.initialized

endproc
wproc( "FBUT_Release_Globals" )

;=================================================== FBUT_Event
proc FBUT_Event
; Put a call to this proc in your WAITPROC or after a GETEVENT.
;
; Pass it any KEY and MOUSE DOWN events *before* your event
; handler looks at them.
;
; If this proc returns -1, then you should let your event handler
; look at the event. -1 means that event was not handled by this
; proc.
;
; This proc only handles KEY and MOUSE events that are associated
; with an active BUTTON, RADIOBUTTON, or CHECKBOX floating window.
;
; The only other values that can be returned are 1 and 2.
;
; The normal return is 1. If you are calling this proc from a
; WAITPROC, then return the 1 to the WAIT so the event will not
; be processed by Paradox. If you are calling from a GETEVENT
; loop, then your event handler should ignore the event.
;
; If 2 is returned and you are calling from a WAITPROC, then you
; should return the 2 to end the WAIT. If you are calling from a
; GETEVENT loop, then you should exit from the loop.
;
; Example of a call in a WAITPROC:
;
; proc wait.proc( trigger.type, event.rec, cycle.number )
;
; <code that doesn't handle key or mouse events>
;
; if trigger.type = "EVENT" and
;    ( event.rec[ "type" ] = "KEY" or
;      ( event.rec[ "type" ] = "MOUSE" and
;        event.rec[ "action" ] = "DOWN" ) ) then
;     if FBUT_Event( event.rec ) > 0 then
;         return retval  ; retval from FBUT_Event
;     endif
; endif
;
; <your event handling code>
;
; endproc
;
;
; Example of a call in a GETEVENT loop:
;
; while <your loop condition>
;     getevent <your event list> to event.rec
;     if event.rec[ "type" ] = "KEY" or
;        ( event.rec[ "type" ] = "MOUSE" and
;          event.rec[ "action" ] = "DOWN" ) then
;         if FBUT_Event( event.rec ) > 0 then
;             if retval = 2 then
;                 quitloop    ; leave the getevent loop
;             else
;                 loop        ; get next event
;             endif
;         endif
;     endif
;     <your event handling code>
; endwhile
;
; Return value:
; -1 if this event is not handled.
; Retval from execproc if it is handled.
; The retval will be forced to 1 or 2.
(
    evt   ; {I} The event record
)

switch
case lower( evt[ "type" ] ) = "key" and
     isassigned( fbut.hotkey.list[ evt[ "keycode" ] ] ) :
    return FBUT_Key_Event( evt )

case lower( evt[ "type" ] ) = "mouse" and
     lower( evt[ "action" ] ) = "down" and
     isassigned(
         fbut.window.parent[ windowat( evt[ "row" ], evt[ "col" ] ) ] ) :
    return FBUT_Mouse_Event( evt )
endswitch

return -1    ; not handled

endproc
wproc( "FBUT_Event" )


;>=================================================================
;>=================================================================
;>
;> The procs that follow are used internally and are of no interest
;> to anyone using a floating BUTTON, RADIOBUTTON, or CHECKBOX
;> window.
;>
;>=================================================================
;>=================================================================
;\

;=================================================== FBUT_Get_Alt_Keycode
proc FBUT_Get_Alt_Keycode
; Return the keycode for Alt-x, where x is the input character.
;
; The input character must be:
;  - A letter (case doesn't matter)
;  - A digit (0-9)
;  - A dash "-"  (Underscore "_" is equivalent to dash.)
;  - An equal sign "=" ( Plus "+" is equivalent to equal.)
;
; blanknum() will be returned for any other character.
;
;
; Return value:
; The Alt keycode or blanknum()
(
    input.char    ; {I} The char for which the caller wants
                  ;     the Alt-char keycode.
                  ;     If input.char contains more than one
                  ;     character, then the first char is
                  ;     used.
)
private x

dynarray x[]
x[ "A" ] = -30
x[ "B" ] = -48
x[ "C" ] = -46
x[ "D" ] = -32
x[ "E" ] = -18
x[ "F" ] = -33
x[ "G" ] = -34
x[ "H" ] = -35
x[ "I" ] = -23
x[ "J" ] = -36
x[ "K" ] = -37
x[ "L" ] = -38
x[ "M" ] = -50
x[ "N" ] = -49
x[ "O" ] = -24
x[ "P" ] = -25
x[ "Q" ] = -16
x[ "R" ] = -19
x[ "S" ] = -31
x[ "T" ] = -20
x[ "U" ] = -22
x[ "V" ] = -47
x[ "W" ] = -17
x[ "X" ] = -45
x[ "Y" ] = -21
x[ "Z" ] = -44

x[ "1" ] = -120
x[ "2" ] = -121
x[ "3" ] = -122
x[ "4" ] = -123
x[ "5" ] = -124
x[ "6" ] = -125
x[ "7" ] = -126
x[ "8" ] = -127
x[ "9" ] = -128
x[ "0" ] = -129

x[ "-" ] = -130
x[ "_" ] = -130

x[ "=" ] = -131
x[ "+" ] = -131

if isblank( input.char ) then
    return blanknum()
else
    if len( input.char ) > 1 then
        input.char = substr( input.char, 1, 1 )
    endif
    if isassigned( x[ input.char ] ) then
        return x[ input.char ]
    else
        return blanknum()
    endif
endif

endproc
wproc( "FBUT_Get_Alt_Keycode" )

;=================================================== FBUT_Extract_Delimiters
proc FBUT_Extract_Delimiters
; Extract the paired delimiters that bracket a substring of the
; input string and return:
;
;  - the input string with the delimiters removed
;  - the start position of the bracketed substring (relative to 1)
;  - the length of the bracketed substring
;  - the substring
;
; The length of the input string must not exceed 255.
;
; If multiple pairs of delimiters are present, only the first pair
; will be extracted. If multiple pairs of delimiters must be
; processed, then code something like:
;
;   z = string.with.delimiters
;   the.delim = <some delimiter>
;   while true
;       dynarray substr.info[]
;       z = FBUT_Extract_Delimiters( z, the.delim, substr.info )
;       if substr.info[ "length" ] <> 0 then
;           <process the substring>
;       else
;           quitloop
;       endif
;   endwhile
;
;
; Return value:
; String with the first pair of delimiters removed.
; Elements are inserted in substr.info with the following indexes:
;
; Index      Description
; ---------- ---------------------------------------
; startpos   start position of bracketed substring
; length     length of substring
; substring  the substring
(
    input.str     ; {I} String to be examined for delimiters
  , delim.char    ; {I} Delimiter character.
                  ;     If omitted, tildes ~~ will be used.
                  ;     Don't use an ampersand as the delimiter.
  , substr.info   ; {O} Dynarray to hold results.
)
private x
    , l.part
    , r.part
    , m.part
    , match.pat

if isblank( delim.char ) then
    delim.char = "~"
endif
if len( delim.char ) > 1 then
    delim.char = substr( delim.char, 1, 1 )
endif
if delim.char = "." then
    delim.char = "\".\""
endif
if delim.char = "@" then
    delim.char = "\"@\""   ; this probably doesn't help
endif

match.pat = ".." + delim.char + ".." + delim.char + ".."
if match( input.str, match.pat, l.part, m.part, r.part ) then
    substr.info[ "startpos" ] = len( l.part ) + 1
    substr.info[ "length" ] = len( m.part )
    substr.info[ "substring" ] = m.part
    input.str = l.part + m.part + r.part
else
    substr.info[ "startpos" ] = 0
    substr.info[ "length" ] = 0
    substr.info[ "substring" ] = 0
endif

return input.str

endproc
wproc( "FBUT_Extract_Delimiters" )

;======================================================= FBUT_Parse_List
proc FBUT_Parse_List
; Parse a delimited list into a sequentially indexed dynarray.
;
;
; Return value:
; None.
; Entries are added to the dynarray x.list.
(
    d.str         ; {I} The delimited list (string or memo var)
  , x.list        ; {O} A dynarray to hold the entries extracted
                  ;     from d.str.
                  ;     x.list must be declared as a dynarray
                  ;     by the caller.
  , delim.char    ; {I} The delimiter character.
                  ;     Defaults to \n if blank.
)
private x, y, z
    , sp
    , ep
    , len.of.str

if isblank( delim.char ) then
    delim.char = "\n"
endif

if len( delim.char ) > 1 then
    delim.char = substr( delim.char, 1, 1 )
endif

x = 0
d.str = d.str + delim.char
len.of.str = len( d.str )
sp = 1
ep = 1
while sp <= len.of.str
    ep = searchfrom( delim.char, d.str, sp )
    x = x + 1
    x.list[ x ] = substr( d.str, sp, ep - sp ) ; don't include delim
    sp = ep + 1
endwhile

endproc
wproc( "FBUT_Parse_List" )

;================================================= FBUT_Register_New_Window
proc FBUT_Register_New_Window
; Save information about a new window by creating entries in
; several global dynarrays.
;
; This information will be used by window drawing and painting
; procs.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
  , wininfo       ; {I} Dynarray containing window info.
)
private x, y, z

fbut.window.parent[ winhand ] = wininfo[ "parent" ]

z = strval( winhand ) + ":"

for x from 1 to arraysize( fbut.winreg )
    if lower( fbut.winreg[ x ] ) = "parent" then
        loop
    endif
    fbut.window.info[ z + fbut.winreg[ x ] ] = wininfo[ fbut.winreg[ x ] ]
endfor

endproc
wproc( "FBUT_Register_New_Window" )

;================================================= FBUT_Get_Window_Info
proc FBUT_Get_Window_Info
; Get window information created by FBUT_Register_New_Window and
; updated by FBUT_Set_Window_Info.
;
; Return value:
; TRUE if the window info exists.
; The wininfo dynarray will be filled.
; Otherwise, FALSE. (The dynarray is unchanged.)
(
    winhand       ; {I} The floating window handle
  , wininfo       ; {I} Dynarray containing window info.
)
private x

; If fbut.window.parent[ winhand] exists, then we assume
; that ALL of the dynarray entries exist. Otherwise, we
; assume that NONE exist.
;
; We don't care about the existence of the window whose
; handle is winhand.

if isassigned( fbut.window.parent[ winhand ] ) then
    wininfo[ "parent" ] = fbut.window.parent[ winhand ]
    z = strval( winhand ) + ":"
    for x from 1 to arraysize( fbut.winreg )
        if lower( fbut.winreg[ x ] ) = "parent" then
            loop
        endif
        wininfo[ fbut.winreg[ x ] ] = fbut.window.info[ z + fbut.winreg[ x ] ]
    endfor
    return true
else
    return false
endif

endproc
wproc( "FBUT_Get_Window_Info" )

;================================================= FBUT_Set_Window_Info
proc FBUT_Set_Window_Info
; Update window information obtained by calling
; the proc FBUT_Get_Window_Info.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
  , wininfo       ; {I} Dynarray containing window info.
)
private x

fbut.window.parent[ winhand ] = wininfo[ "parent" ]

z = strval( winhand ) + ":"
for x from 1 to arraysize( fbut.winreg )
    if lower( fbut.winreg[ x ] ) = "parent" then
        loop
    endif
    fbut.window.info[ z + fbut.winreg[ x ] ] = wininfo[ fbut.winreg[ x ] ]
endfor

endproc
wproc( "FBUT_Set_Window_Info" )

;==================================================== FBUT_Store_Label_Info
proc FBUT_Store_Label_Info
; Store information about a label.
;
; This information will be retrieved by FBUT_Retrieve_Label_Info.
;
; The indexes in label.info are described below.
;
; Index       Description
; -----       -------------------------------------------------
; ROW         Origin row of label in its window
; COL         Origin col of label in its window
; WIDTH       Width of label column or width of window if labels
;             are in a single column.
; HOTPOS      Offset (relative to 1) of the highlighted area
;             in the label. Usually this is a string that
;             identifies the hotkey used to "select" the label.
; HOTWID      Width of the highlighted area.
; KEYCODE     The keycode of the hotkey associated with the label.
;             If there is no hotkey for the label, then this will
;             be BLANKNUM().
; TEXT        For a BUTTON, this is the text of the label. This
;             is the value placed in the VARNAME field when the
;             button is pushed.
;             For other windows, this field is blank.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
  , label.num     ; {I} The label number within the window
  , label.info    ; {I} Dynarray containing label info
)
private x, y, z

x = strval( winhand ) + ":" + strval( label.num ) + ":"
for y from 1 to arraysize( fbut.labelreg )
    fbut.label.info[ x + fbut.labelreg[ y ] ] =
        label.info[ fbut.labelreg[ y ] ]
endfor

endproc
wproc( "FBUT_Store_Label_Info" )

;================================================== FBUT_Retrieve_Label_Info
proc FBUT_Retrieve_Label_Info
;
; See FBUT_Store_Label_Info for a description of the indexes in
; label.info.
;
; Return value:
; TRUE if the label info exists. The label.info dynarray is filled.
; FALSE if the label info doesn't exist. The dynarray is unchanged.
(
    winhand       ; {I} The floating window handle
  , label.num     ; {I} The label number within the window
  , label.info    ; {I} Dynarray containing label info
)
private x

x = strval( winhand ) + ":" + strval( label.num ) + ":"
if isassigned( fbut.label.info[ x + "row" ] ) then
    for y from 1 to arraysize( fbut.labelreg )
        label.info[ fbut.labelreg[ y ] ] =
            fbut.label.info[ x + fbut.labelreg[ y ] ]
    endfor
    return true
else
    return false
endif

endproc
wproc( "FBUT_Retrieve_Label_Info" )

;================================================== FBUT_Get_Label_Keycode
proc FBUT_Get_Label_Keycode
;
; See FBUT_Store_Label_Info for a description of the indexes in
; label.info.
;
; Return value:
; The keycode of the hotkey associated with the label or
; blanknum() if the label doesn't exist.
(
    winhand       ; {I} The floating window handle
  , label.num     ; {I} The label number within the window
)
private x

x = strval( winhand ) + ":" + strval( label.num ) + ":"
if isassigned( fbut.label.info[ x + "keycode" ] ) then
    return fbut.label.info[ x + "keycode" ]
else
    return blanknum()
endif

endproc
wproc( "FBUT_Get_Label_Keycode" )

;==================================================== FBUT_Hotkey_Active
proc FBUT_Hotkey_Active
; Activate or deactivate the hotkey for a label
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
  , label.num     ; {I} The label number within the window
  , x.active      ; {I} TRUE->The hotkey is activated.
                  ;     FALSE->The hotkey is deactivated.
)
private x
  , hotkey.keycode

if iswindow( winhand ) then
    hotkey.keycode = FBUT_Get_Label_Keycode( winhand, label.num )
    if not isblank( hotkey.keycode ) then
        x = strval( winhand ) + ":" + strval( label.num )
        if x.active then
            fbut.hotkey.list[ hotkey.keycode ] = x
        else
            ; We only remove the hotkey from the list if it's
            ; for the winhand window. The caller may have
            ; activated some other window using the same hotkey
            ; before deactivating the winhand window.
            if isassigned( fbut.hotkey.list[ hotkey.keycode ] ) and
               fbut.hotkey.list[ hotkey.keycode ] = x then
                release vars fbut.hotkey.list[ hotkey.keycode ]
            endif
        endif
    endif
endif

endproc
wproc( "FBUT_Hotkey_Active" )

;=================================================== FBUT_Key_Event
proc FBUT_Key_Event
; Called by FBUT_Event to handle a KEY event.
;
; The fact that this proc has been called means that the
; event's keycode is a hotkey for an active window.
;
;
; Return value:
; -1 if this event is not handled.
; Retval from execproc if it is handled.
; The retval will be forced to 1 or 2.
(
    evt   ; {I} The KEY event record
)
private x, y, z
    , label.id
    , label.num
    , label.info
    , winhand
    , wininfo
    , r.val

if not isassigned( fbut.hotkey.list[ evt[ "keycode" ] ] ) then
    return -1   ; we got here by mistake
endif

r.val = 1  ; default return value

label.id = fbut.hotkey.list[ evt[ "keycode" ] ]

; label.id = winhand : label.num
if match( label.id, "..:..", winhand, label.num ) then
    winhand = numval( winhand )
    label.num = numval( label.num )
    dynarray wininfo[]      ; Globals for FBUT_Set_User_Var
    dynarray label.info[]
    if FBUT_Get_Window_Info( winhand, wininfo ) and
       FBUT_Retrieve_Label_Info( winhand, label.num, label.info ) then
        r.val = FBUT_Set_User_Var( winhand, label.num )
    endif
endif

return r.val

endproc
wproc( "FBUT_Key_Event" )

;=================================================== FBUT_Mouse_Event
proc FBUT_Mouse_Event
; Called by FBUT_Event to handle a mouse event.
;
; The fact that this proc was called means that a mouse down
; occurred on a floating BUTTON, etc. window.
;
;
; Return value:
; -1 if this event is not handled.
; Retval from execproc if it is handled.
; The retval will be forced to 1 or 2.
(
    evt   ; {I} The MOUSE event record
)
private x, y, z
    , x.row
    , x.col
    , r.val
    , winhand
    , label.info
    , wininfo

localizeevent evt
winhand = evt[ "window" ]
if not isassigned( fbut.window.parent[ winhand ] ) then
    globalizeevent evt
    return -1   ; we got here by mistake
endif

; Currently, none of the windows are framed. but future windows
; may be framed. We don't want to process a mouse event that is
; on the frame.

if isassigned( evt[ "onframe" ] ) and
   evt[ "onframe" ] then
    globalizeevent evt
    return -1   ; let caller or Paradox process this event
endif

x.row = evt[ "row" ]
x.col = evt[ "col" ]

window getattributes winhand to x

; If our coordinates are on the frame, it means the
; mouse down occurred on the frame.
; This can happen in 4.0x because the "ONFRAME" index is
; not supplied by localizeevent.

if x[ "hasframe" ] then
    if x.row = 0 or
       x.row = x[ "width" ] or
       x.col = 0 or
       x.col = x[ "height" ] then
        globalizeevent evt
        return -1   ; let caller or Paradox process this event
    endif
endif

; If the window is framed, we want to subtract 1 from the
; row and col because Paradox passes the event with 0, 0
; as the upper left corner of the frame.

if x[ "hasframe" ] then
    x.row = x.row - 1
    x.col = x.col - 1
endif

; The event is on "our" window. We'll wait for the mouse up event
; and discard it.
getevent mouse "UP" to x
release vars x

; *** FROM THIS POINT ON, -1 IS NOT A VALID RETURN VALUE. ***

r.val = 1    ; set default return value

dynarray wininfo[]
dynarray label.info[]

if FBUT_Get_Window_Info( winhand, wininfo ) then
    ; We want to find the label that the mouse is sitting on
    for x from 1 to wininfo[ "numlbl" ]
        if FBUT_Retrieve_Label_Info( winhand, x, label.info ) then
            if x.row = label.info[ "row" ] and
               x.col >= label.info[ "col" ] and
               x.col < label.info[ "col" ] + label.info[ "width" ] then
                quitloop
            endif
            ; If we're on a BUTTON window, any click on the window
            ; applies to the label. (This code is here to handle
            ; BUTTONS with internal frames, etc. )
            if wininfo[ "type" ] = "B" then
                quitloop
            endif
        endif
    endfor
    if x <= wininfo[ "numlbl" ] then   ; we found the label
        r.val = FBUT_Set_User_Var( winhand, x )
    else
        ; The mouse was clicked in the window but not on one
        ; of the labels. We ignore the click.
        ; r.val is already set to 1.
    endif
endif

return r.val

endproc
wproc( "FBUT_Mouse_Event" )

;======================================================= FBUT_Set_User_Var
proc FBUT_Set_User_Var
; Internal proc to set the user var associated with a
; label and to execute the proc associated with the
; label's window.
;
; The caller provides the following global dynarrays:
;
;  - wininfo     - information about the window
;  - label.info  - information about the label
;
; Return value:
; The return value from the user's proc.
; The only valid values are 1 and 2.
; Any other value is forced to 1.
(
    winhand        ; {I} The handle of the window to process
  , label.num      ; {I} The label number within the window
)
private x, y, z
    , x.varname
    , x.procname
    , r.val

r.val = 1  ; default return value

x.procname = wininfo[ "procname" ]
x.varname = wininfo[ "varname" ]

if not isblank( x.varname ) then
    switch
    case wininfo[ "type" ] = "B" :       ; BUTTON
        varassign x.varname from label.info[ "text" ]

    case wininfo[ "type" ] = "C" :      ; CHECKBOX
        ; The var for a checkbox is a dynarray
        ; indexed by label number.
        ; The value is a boolean that must be toggled.
        z = not varval( x.varname, label.num )
        varassign x.varname index label.num from z

    case wininfo[ "type" ] = "R" :      ; RADIOBUTTON
        varassign x.varname from label.num

    endswitch
endif

if not isblank( x.procname ) then

    ; If x.procname doesn't exist, an error will occur.
    ; We don't use ISPROC() because we want the error to
    ; occur so the problem can be detected and fixed.

    execproc x.procname
    r.val = retval

    ; Force r.val to 1 or 2
    if type( r.val ) <> "N" then
        r.val = 1
    endif
    if r.val <> 1 and r.val <> 2 then
        r.val = 1
    endif
endif

FBUT_Update( winhand )

return r.val

endproc
wproc( "FBUT_Set_User_Var" )

;=================================================== FBUT_Paint
proc FBUT_Paint
; Internal proc to paint a floating window.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
)
private x, y, z
    , orig.canv
    , wininfo
    , winattr

; Make sure the window exists
if iswindow( winhand ) then
    dynarray wininfo[]
    if FBUT_Get_Window_Info( winhand, wininfo ) then
        orig.canv = getcanvas()
        setcanvas winhand
        window getattributes winhand to winattr

        execproc fbut.paint.proc[ wininfo[ "type" ] ]

        if iswindow( orig.canv ) then
            setcanvas orig.canv
        else
            setcanvas default
        endif
    endif
endif

endproc
wproc( "FBUT_Paint" )

;=================================================== FBUT_Update
proc FBUT_Update
; Internal proc to update the "selected" marks in a
; floating window.
;
; Return value:
; None.
(
    winhand       ; {I} The floating window handle
)
private x, y, z
    , orig.canv
    , wininfo
    , winattr

; Make sure the window exists
if iswindow( winhand ) then
    dynarray wininfo[]
    if FBUT_Get_Window_Info( winhand, wininfo ) then
        orig.canv = getcanvas()
        setcanvas winhand
        window getattributes winhand to winattr

        execproc fbut.update.proc[ wininfo[ "type" ] ]

        if iswindow( orig.canv ) then
            setcanvas orig.canv
        else
            setcanvas default
        endif
    endif
endif

endproc
wproc( "FBUT_Update" )


;>================================================================
;>================================================================
;>
;> Procs for floating BUTTON windows.
;>
;>================================================================
;>================================================================
;\


;=================================================== FBUT_Draw_Button
proc FBUT_Draw_Button
; Internal proc to create the floating window for a button and
; put text in the window.
;
; The button is not painted. (See FBUT_Paint_Button)
;
; The caller must provide the global dynarrays:
;  - wininfo
;  - x.label
;  - x.keycode
;  - x.hipos
;  - x.hilen
;
; These vars are described in the proc FBUT_Create.
;
;
; Return value:
; The button window's handle.
(
)
private x, y, z
    , winhand
    , left.pad
    , label.info

switch
case wininfo[ "style" ] = 1 :
    ; 1 row high. No frame. No shadow.

    wininfo[ "height" ] = 1
    wininfo[ "width" ] =
        max( wininfo[ "maxwid" ] + 2, wininfo[ "width" ] )

    ; Set up attributes for style 1
    dynarray x[]
    x[ "canclose" ] = false
    x[ "canmaximize" ] = false
    x[ "canmove" ] = false
    x[ "canresize" ] = false
    x[ "canvasheight"] = wininfo[ "height" ]
    x[ "canvaswidth" ] = wininfo[ "width" ]
    x[ "floating" ] = true
    x[ "hasframe" ] = false
    x[ "hasshadow" ] = false
    x[ "height" ] = wininfo[ "height" ]
    x[ "maximized" ] = false
    x[ "origincol" ] = wininfo[ "hidecol" ]
    x[ "originrow" ] = wininfo[ "hiderow" ]
    x[ "width" ] = wininfo[ "width" ]

    window create attributes x to winhand
    release vars x
    setcanvas winhand

    ; Center label in window
    left.pad = int( ( wininfo[ "width" ] - len( x.label[ 1 ] ) ) / 2 )
    ?? spaces( left.pad ) + x.label[ 1 ]
    dynarray label.info[]
    label.info[ "row" ] = 0
    label.info[ "col" ] = 0
    label.info[ "width" ] = wininfo[ "width" ]
    if x.hipos[ 1 ] > 0 then
        label.info[ "hotpos" ] = x.hipos[ 1 ] + left.pad
        label.info[ "hotwid" ] = x.hilen[ 1 ]
    else
        label.info[ "hotpos" ] = 0
        label.info[ "hotwid" ] = 0
    endif
    label.info[ "keycode" ] = x.keycode[ 1 ]
    label.info[ "text" ] = x.label[ 1 ]
    FBUT_Store_Label_Info( winhand, 1, label.info )

otherwise :
    ; invalid button style - force to 1
    wininfo[ "style" ] = 1
    winhand = FBUT_Draw_Button()
endswitch

return winhand

endproc
wproc( "FBUT_Draw_Button" )

;=================================================== FBUT_Paint_Button
proc FBUT_Paint_Button
; Internal proc to paint the floating window for a button.
;
; Called by FBUT_Paint with the following global vars:
;
;   winhand       - window handle
;   winattr       - window attributes (from getattributes)
;   wininfo       - dynarray of window info
;
; Return value:
; None.
(
)
private x, y, z
    , window.color
    , hotkey.color
    , hotkey.row
    , hotkey.col
    , hotkey.wid
    , label.info

switch
case wininfo[ "style" ] = 1 :
    if wininfo[ "active" ] then
        window.color = fbut.color.button.window.active
        hotkey.color = fbut.color.button.hotkey.active
    else
        window.color = fbut.color.button.window.inactive
        hotkey.color = fbut.color.button.hotkey.inactive
    endif
    paintcanvas attribute window.color all
    dynarray label.info[]
    if FBUT_Retrieve_Label_Info( winhand, 1, label.info ) then
        ; If hotpos is zero, there is no highlight for
        ; this label.
        if label.info[ "hotpos" ] > 0 then
            hotkey.row = label.info[ "row" ]
            hotkey.col = label.info[ "hotpos" ] + label.info[ "col" ] - 1
            hotkey.wid = label.info[ "hotwid" ]
            z = hotkey.col + hotkey.wid - 1  ; end col for paint
            paintcanvas attribute hotkey.color
                hotkey.row, hotkey.col, hotkey.row, z
        endif
    endif

otherwise :
    ; invalid button style
endswitch

endproc
wproc( "FBUT_Paint_Button" )

;=================================================== FBUT_Update_Button
proc FBUT_Update_Button
; Internal proc to update the "selected" marks in the floating
; window for a button.
;
; The caller must supply the following global vars:
;
;   winhand       - window handle
;   winattr       - window attributes (from getattributes)
;   wininfo       - dynarray of window info
;
;
; Return value:
; None.
(
)
private x

; Since there are no marks, this proc returns immediately.

endproc
wproc( "FBUT_Update_Button" )

;>================================================================
;>================================================================
;>
;> Procs for floating CHECKBOX windows.
;>
;>================================================================
;>================================================================
;\

;=================================================== FBUT_Draw_CheckBox
proc FBUT_Draw_CheckBox
; Internal proc to create the floating window for a checkbox and
; put text in the window.
;
; The checkbox is not painted. (See FBUT_Paint_CheckBox)
;
; The caller must provide the global dynarrays:
;  - wininfo
;  - x.label
;  - x.keycode
;  - x.hipos
;  - x.hilen
;
; These vars are described in the proc FBUT_Create.
;
;
; Return value:
; The window handle.
(
)
private x, y, z
    , winhand
    , label.info
    , t.hgt

switch
case wininfo[ "style" ] = 1 :
    ; No frame. No shadow.
    ; Set up attributes for style 1

    ; Leave room for [ ] and margins
    wininfo[ "width" ] = max( wininfo[ "maxwid" ] + 6, wininfo[ "width" ] )

    ; Height is number of labels + heading
    if wininfo[ "hasheading" ] then
        wininfo[ "height" ] = wininfo[ "numlbl" ] + 1
    else
        wininfo[ "height" ] = wininfo[ "numlbl" ]
    endif

    dynarray x[]
    x[ "canclose" ] = false
    x[ "canmaximize" ] = false
    x[ "canmove" ] = false
    x[ "canresize" ] = false
    x[ "canvasheight"] = wininfo[ "height" ]
    x[ "canvaswidth" ] = wininfo[ "width" ]
    x[ "floating" ] = true
    x[ "hasframe" ] = false
    x[ "hasshadow" ] = false
    x[ "height" ] = wininfo[ "height" ]
    x[ "maximized" ] = false
    x[ "origincol" ] = wininfo[ "hidecol" ]
    x[ "originrow" ] = wininfo[ "hiderow" ]
    x[ "width" ] = wininfo[ "width" ]

    window create attributes x to winhand
    release vars x
    setcanvas winhand

    if wininfo[ "hasheading" ] then
        @ 0,1 ?? wininfo[ "heading" ]
        t.hgt = 1
    else
        t.hgt = 0
    endif

    dynarray label.info[]
    for x from 1 to wininfo[ "numlbl" ]
        @ t.hgt + x - 1, 0
        ?? " [ ] " + x.label[ x ]
        label.info[ "row" ] = t.hgt + x - 1
        label.info[ "col" ] = 0
        label.info[ "width" ] = wininfo[ "width" ]
        if x.hipos[ x ] > 0 then
            label.info[ "hotpos" ] = x.hipos[ x ] + 5
            label.info[ "hotwid" ] = x.hilen[ x ]
        else
            label.info[ "hotpos" ] = 0
            label.info[ "hotwid" ] = 0
        endif
        label.info[ "keycode" ] = x.keycode[ x ]
        label.info[ "text" ] = ""  ; no need to save label
        FBUT_Store_Label_Info( winhand, x, label.info )
    endfor

otherwise :
    ; invalid style - force to 1
    wininfo[ "style" ] = 1
    winhand = FBUT_Draw_CheckBox()
endswitch

return winhand

endproc
wproc( "FBUT_Draw_CheckBox" )

;=================================================== FBUT_Paint_CheckBox
proc FBUT_Paint_CheckBox
; Internal proc to paint the floating window for a checkbox.
;
; Called by FBUT_Paint with the following global vars:
;
;   winhand       - window handle
;   winattr       - window attributes (from getattributes)
;   wininfo       - dynarray of window info
;
;
; Return value:
; None.
(
)
private x, y, z
    , label.info
    , window.color
    , title.color
    , hotkey.color
    , hotkey.row
    , hotkey.col
    , hotkey.wid

switch
case wininfo[ "style" ] = 1 :
    if wininfo[ "active" ] then
        window.color = fbut.color.check.window.active
        title.color  = fbut.color.check.title.active
        hotkey.color = fbut.color.check.hotkey.active
    else
        window.color = fbut.color.check.window.inactive
        title.color  = fbut.color.check.title.inactive
        hotkey.color = fbut.color.check.hotkey.inactive
    endif
    paintcanvas attribute window.color all
    if wininfo[ "hasheading" ] then
        paintcanvas attribute title.color 0, 0, 0, winattr[ "width" ] - 1
    endif
    dynarray label.info[]
    for x from 1 to wininfo[ "numlbl" ]
        ; If hotpos is zero, there is no highlight for this label.
        if FBUT_Retrieve_Label_Info( winhand, x, label.info ) and
           label.info[ "hotpos" ] > 0 then
            hotkey.row = label.info[ "row" ]
            hotkey.col = label.info[ "hotpos" ] + label.info[ "col" ] - 1
            hotkey.wid = label.info[ "hotwid" ]
            z = hotkey.col + hotkey.wid - 1  ; end col for paint
            paintcanvas attribute hotkey.color
                hotkey.row, hotkey.col, hotkey.row, z
        endif
    endfor

otherwise :
    ; invalid style
endswitch

endproc
wproc( "FBUT_Paint_CheckBox" )

;=================================================== FBUT_Update_CheckBox
proc FBUT_Update_CheckBox
; Internal proc to update the "selected" marks in the floating
; window for a checkbox.
;
; The caller must supply the following global vars:
;
;   winhand       - window handle
;   winattr       - window attributes (from getattributes)
;   wininfo       - dynarray of window info
;
;
; Return value:
; None.
(
)
private x, y, z
    , label.info
    , var.val

switch
case wininfo[ "style" ] = 1 :
    if wininfo[ "active" ] then
        style attribute fbut.color.check.window.active
    else
        style attribute fbut.color.check.window.inactive
    endif
    dynarray label.info[]
    for x from 1 to wininfo[ "numlbl" ]
        if FBUT_Retrieve_Label_Info( winhand, x, label.info ) then
            if isblank( wininfo[ "varname" ] ) then
                var.val = false
            else
                var.val = varval( wininfo[ "varname" ], x )
            endif

            @ label.info[ "row" ], label.info[ "col" ] + 2
            if var.val then
                ?? chr( 251 )  ; root symbol - looks like checkmark
            else
                ?? " "
            endif
        endif
    endfor

otherwise :
    ; invalid style
endswitch

endproc
wproc( "FBUT_Update_CheckBox" )


;>================================================================
;>================================================================
;>
;> Procs for floating RADIOBUTTON windows.
;>
;>================================================================
;>================================================================
;\


;=================================================== FBUT_Draw_Radio
proc FBUT_Draw_Radio
; Internal proc to create the floating window for a radiobutton and
; put text in the window.
;
; The radiobutton is not painted. (See FBUT_Paint_Radio)
;
; The caller must provide the global dynarrays:
;  - wininfo
;  - x.label
;  - x.keycode
;  - x.hipos
;  - x.hilen
;
; These vars are described in the proc FBUT_Create.
;
;
; Return value:
; The window handle.
(
)
private x, y, z
    , label.info
    , winhand
    , t.hgt

switch
case wininfo[ "style" ] = 1 :
    ; No frame. No shadow.
    ; Set up attributes for style 1

    ; Leave room for ( ) and margins
    wininfo[ "width" ] = max( wininfo[ "maxwid" ] + 6, wininfo[ "width" ] )

    ; Height is number of labels + heading
    if wininfo[ "hasheading" ] then
        wininfo[ "height" ] = wininfo[ "numlbl" ] + 1
    else
        wininfo[ "height" ] = wininfo[ "numlbl" ]
    endif

    dynarray x[]
    x[ "canclose" ] = false
    x[ "canmaximize" ] = false
    x[ "canmove" ] = false
    x[ "canresize" ] = false
    x[ "canvasheight"] = wininfo[ "height" ]
    x[ "canvaswidth" ] = wininfo[ "width" ]
    x[ "floating" ] = true
    x[ "hasframe" ] = false
    x[ "hasshadow" ] = false
    x[ "height" ] = wininfo[ "height" ]
    x[ "maximized" ] = false
    x[ "origincol" ] = wininfo[ "hidecol" ]
    x[ "originrow" ] = wininfo[ "hiderow" ]
    x[ "width" ] = wininfo[ "width" ]

    window create attributes x to winhand
    release vars x
    setcanvas winhand

    if wininfo[ "hasheading" ] then
        @ 0,1 ?? wininfo[ "heading" ]
        t.hgt = 1
    else
        t.hgt = 0
    endif

    dynarray label.info[]
    for x from 1 to wininfo[ "numlbl" ]
        @ t.hgt + x - 1, 0
        ?? " ( ) " + x.label[ x ]
        label.info[ "row" ] = t.hgt + x - 1
        label.info[ "col" ] = 0
        label.info[ "width" ] = wininfo[ "width" ]
        if x.hipos[ x ] > 0 then
            label.info[ "hotpos" ] = x.hipos[ x ] + 5
            label.info[ "hotwid" ] = x.hilen[ x ]
        else
            label.info[ "hotpos" ] = 0
            label.info[ "hotwid" ] = 0
        endif
        label.info[ "keycode" ] = x.keycode[ x ]
        label.info[ "text" ] = ""  ; no need to save label
        FBUT_Store_Label_Info( winhand, x, label.info )
    endfor

otherwise :
    ; invalid style - force to 1
    wininfo[ "style" ] = 1
    winhand = FBUT_Draw_Radio()
endswitch

return winhand

endproc
wproc( "FBUT_Draw_Radio" )

;=================================================== FBUT_Paint_Radio
proc FBUT_Paint_Radio
; Internal proc to paint the floating window for a radiobutton.
;
; Called by FBUT_Paint with the following global vars:
;
;   winhand       - window handle
;   winattr       - window attributes (from getattributes)
;   wininfo       - dynarray of window info
;
;
; Return value:
; None.
(
)
private x, y, z
    , label.info
    , window.color
    , title.color
    , hotkey.color
    , hotkey.row
    , hotkey.col
    , hotkey.wid

switch
case wininfo[ "style" ] = 1 :
    if wininfo[ "active" ] then
        window.color = fbut.color.radio.window.active
        title.color  = fbut.color.radio.title.active
        hotkey.color = fbut.color.radio.hotkey.active
    else
        window.color = fbut.color.radio.window.inactive
        title.color  = fbut.color.radio.title.inactive
        hotkey.color = fbut.color.radio.hotkey.inactive
    endif
    paintcanvas attribute window.color all
    if wininfo[ "hasheading" ] then
        paintcanvas attribute title.color 0, 0, 0, winattr[ "width" ] - 1
    endif
    dynarray label.info[]
    for x from 1 to wininfo[ "numlbl" ]
        ; If hotpos is zero, there is no highlight for this label.
        if FBUT_Retrieve_Label_Info( winhand, x, label.info ) and
           label.info[ "hotpos" ] > 0 then
            hotkey.row = label.info[ "row" ]
            hotkey.col = label.info[ "hotpos" ] + label.info[ "col" ] - 1
            hotkey.wid = label.info[ "hotwid" ]
            z = hotkey.col + hotkey.wid - 1  ; end col for paint
            paintcanvas attribute hotkey.color
                hotkey.row, hotkey.col, hotkey.row, z
        endif
    endfor

otherwise :
    ; invalid style
endswitch

endproc
wproc( "FBUT_Paint_Radio" )

;=================================================== FBUT_Update_Radio
proc FBUT_Update_Radio
; Internal proc to update the "selected" marks in the floating
; window for a radiobutton.
;
; The caller must supply the following global vars:
;
;   winhand       - window handle
;   winattr       - window attributes (from getattributes)
;   wininfo       - dynarray of window info
;
;
; Return value:
; None.
(
)
private x, y, z
    , label.info
    , var.val

if isblank( wininfo[ "varname" ] ) then
    var.val = blanknum()
else
    var.val = varval( wininfo[ "varname" ] )
endif

switch
case wininfo[ "style" ] = 1 :
    if wininfo[ "active" ] then
        style attribute fbut.color.radio.window.active
    else
        style attribute fbut.color.radio.window.inactive
    endif
    dynarray label.info[]
    for x from 1 to wininfo[ "numlbl" ]
        if FBUT_Retrieve_Label_Info( winhand, x, label.info ) then
            @ label.info[ "row" ], label.info[ "col" ] + 2
            if x = var.val then
                ?? chr( 7 )  ; bullet
            else
                ?? " "
            endif
        endif
    endfor

otherwise :
    ; invalid style
endswitch

endproc
wproc( "FBUT_Update_Radio" )

